#--------------------------------------------------------------------------
#     This file is part of BKchem - a chemical drawing program
#     Copyright (C) 2002, 2003 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------
#
# Last edited: $Date: 2003/09/26 19:30:41 $
#
#--------------------------------------------------------------------------

"""this module provides basic coordinate transformations in 3D based on matrix algebra"""

from math import cos, sin, pi

class transform3d:
  """this class provides basic higher-level interface for coordinate transforms"""
  def __init__( self, mat = None):
    if not mat:
      self.mat = matrix( [[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]])
    else:
      self.mat = matrix( mat)

  def transform_xyz( self, x, y, z):
    x1, y1, z1, one = self.mat.get_multiplied2( [[x],[y],[z],[1]])
    return x1[0], y1[0], z1[0]

  def transform_list( self, l):
    ret = []
    for line in l:
      ret.append( self.transform_xyz( line[0], line[1], line[2]))
    return ret

  def map_transform( self, l):
    "this method can be used with map() function"
    return self.transform_xyz( l[0], l[1], l[2])

  def set_move( self, dx, dy, dz):
    "add an moving step to transformation matrix"
    self.mat = matrix( mat=self.mat.get_multiplied( [[1,0,0,dx],[0,1,0,dy],[0,0,1,dz],[0,0,0,1]]))

  def set_rotation( self, xa, ya, za):
    "add an rotation step to transformation matrix"
    self.set_rotation_x( xa)
    self.set_rotation_y( ya)
    self.set_rotation_z( za)
    
  def set_rotation_x( self, xa):
    mat = [[1,0,0,0],
           [0, cos(xa), sin(xa), 0],
           [0, -sin(xa), cos(xa), 0],
           [0,0,0,1]]
    self.mat = matrix( mat=self.mat.get_multiplied( mat))

  def set_rotation_y( self, ya):
    mat = [[cos(ya), 0, -sin(ya), 0],
           [0, 1, 0, 0],
           [sin(ya), 0, cos(ya), 0],
           [0,0,0,1]]
    self.mat = matrix( mat=self.mat.get_multiplied( mat))

  def set_rotation_z( self, za):
    mat = [[cos(za), sin(za), 0, 0],
           [-sin(za), cos(za), 0, 0],
           [0,0,1,0],
           [0,0,0,1]]
    self.mat = matrix( mat=self.mat.get_multiplied( mat))

  def set_scaling_xyz( self, mx, my, mz):
    "add an scaling step to transformation matrix"
    self.mat = matrix( mat=self.mat.get_multiplied( [[mx,0,0,0],[0,my,0,0],[0,0,mz,0],[0,0,0,1]]))

  def set_scaling( self, scale):
    "add an scaling step to transformation matrix, same scaling for both dimensions"
    self.set_scaling( scale, scale, scale)



class matrix:
  "this class provides common operations for matrix of 3x3 elements"

  def __init__( self, mat):
    self.mat = mat

  def get_transposed( self):
    m = self.mat
    ret = [[0,0,0,0],[0,0,0,0],[0,0,0,0],[0,0,0,0]]
    for i in range( len( m)):
      for j in range( len( m)):
        ret[j][i] = m[i][j]
    return ret

  def get_inverse( self):
    pass

  def get_multiplied2( self, multi):
    "returns matrix multiplied by multi"
    m = []
    if type( multi) != type([]):
      for i in range( len( self.mat)):
        m.append([])
        for j in range( len( self.mat)):
          m[i].append( multi*self.mat[i][j])
    else:
      for i in range( len( self.mat)):
        m.append([])
        for j in range( len( multi[0])):
          m[i].append( self.mat[i][0]*multi[0][j] + self.mat[i][1]*multi[1][j] + self.mat[i][2]*multi[2][j] + self.mat[i][3]*multi[3][j])
    return  m

  def get_multiplied( self, multi):
    "returns matrix multiplied by multi"
    m = []
    if type( multi) != type([]):
      for i in range( len( self.mat)):
        m.append([])
        for j in range( len( self.mat)):
          m[i].append( multi*self.mat[i][j])
    else:
      for i in range( len( multi)):
        m.append([])
        for j in range( len( self.mat)):
          m[i].append( multi[i][0]*self.mat[0][j] + multi[i][1]*self.mat[1][j] + multi[i][2]*self.mat[2][j] +  multi[i][3]*self.mat[3][j])
    return  m


