#--------------------------------------------------------------------------
#     This file is part of BKchem - a chemical drawing program
#     Copyright (C) 2002, 2003 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------
#
# Last edited: $Date: 2003/09/26 19:30:41 $
#
#--------------------------------------------------------------------------

"""this module implements undo_manager and state_record classes,
should provide everything needed to perform undo management.
state_record is used only inside of undo_manager and is not
exported in __all__."""

## NOTE that undo uses a low-level access to objects in order to
## speed up the task.

import misc
import copy
import data

__all__= ['undo_manager']

##-------------------- UNDO MANAGER --------------------

class undo_manager:
  """class to process requests for undo tracking and undoing"""

  MAX_RECORDS = 20
  
  def __init__( self, paper):
    """well, init"""
    self.paper = paper
    self._records = []
    self.clean()
    self.start_new_record()
    
  def start_new_record( self, name=''):
    """starts new undo_record closing the recent
    name may be set for a record"""
    if len( self._records)-1 > self._pos:
      del self._records[ (self._pos+1):]
    if len( self._records) >= self.MAX_RECORDS:
      del self._records[0]
      self._pos -= 1
    self._records.append( state_record( self.paper, name=name))
    self._pos += 1

  def undo( self):
    """undoes the last step and returns the number of undo records available"""
    self._pos -= 1
    if self._pos >= 0:
      self._records[ self._pos].undo( self._records[ self._pos+1])
    else:
      self._pos = 0
    return self._pos

  def redo( self):
    """redoes the last undone step, returns number of redos available"""
    self._pos += 1
    if self._pos < len( self._records):
      self._records[ self._pos].undo( self._records[ self._pos-1])
    else:
      self._pos = len( self._records)-1
    return len( self._records) - self._pos -1

  def clean( self):
    """removes all undo informations, does not start new undo record"""
    self._pos = -1
    del self._records[:]

  def get_last_record_name( self):
    """returns the last closed record name"""
    if self._pos >= 1:
      return self._records[ self._pos-1].name
    else:
      return None

  def delete_last_record( self):
    """deletes the last record, useful for concatenation of several records to one;
    especially powerfull in combination with named records;
    use with care - it could cause problems"""
    if self._pos > 0:
      del self._records[ self._pos-1]
      self._pos -= 1

##-------------------- STATE RECORD --------------------
    
class state_record:
  """class able to store state of the whole system and then set
  the system back to that stored state"""

  def __init__( self, paper, name=''):
    """hmmm, what is supposed to be in comment for __init__?"""
    self.paper = paper
    self.objects = []
    self.records = []
    self.name = name
    self.stack = []
    self.record_state()

  def record_state( self):
    """stores all necessary information about the system, so that its than able to
    fully recover that state."""
    self.stack = copy.copy( self.paper.stack)
    for mol in self.paper.molecules:
      self.objects.append( mol)
      self.records.append( self._mol_rec( mol))
      for o in mol.atoms_map:
        self.objects.append( o)
        self.records.append( self._atom_rec( o))
        for m in o.marks.itervalues():
          if m:
            self.objects.append( m)
            self.records.append( self._mark_rec( m))
      for o in mol.bonds:
        self.objects.append( o)
        self.records.append( self._bond_rec( o))
    for arr in self.paper.arrows:
      self.objects.append( arr)
      self.records.append( self._arrow_rec( arr))
      for o in arr.points:
        self.objects.append( o)
        self.records.append( self._point_rec( o))
    for plus in self.paper.pluses:
      self.objects.append( plus)
      self.records.append( self._plus_rec( plus))
    for text in self.paper.texts:
      self.objects.append( text)
      self.records.append( self._text_rec( text))
    for v in self.paper.vectors:
      self.objects.append( v)
      if v.object_type == 'polygon':
        self.records.append( self._polygon_rec( v))
        for o in v.points:
          self.objects.append( o)
          self.records.append( self._point_rec( o))
      else:
        self.records.append( self._vector_rec( v))

  def undo( self, previous):
    """does undo, actually only calls self.set_state"""
    self.set_state( previous)

  def set_state( self, previous):
    """sets the system to the recorded state (update is done only where necessary,
    not changed values are not touched)."""
    # we need to know about deleted bonds before we try to redraw them (when updating atom)
    deleted = misc.difference( self.objects, previous.objects)
    ## CHANGED OBJECTS
    i = 0
    for o in self.objects:
      changed = 0
      ## ARROW
      if o.object_type == 'arrow':
        for key in ('pin', 'spline', 'line_width', 'line_color'):
          if self.records[i][key] != o.__dict__[key]:
            o.__dict__[key] = self.records[i][key]
            changed = 1
        for key in ('points',):
          if self.records[i][key] != o.__dict__[key]:
            o.__dict__[ key] = copy.copy( self.records[i][key])
            changed = 1
      ## MOLECULE
      elif o.object_type == 'molecule':
        for key in ('name', 'id'):
          if self.records[i][key] != o.__dict__[key]:
            o.__dict__[key] = self.records[i][key]
        for key in ('atoms_map', 'bonds'):
          o.__dict__[ key] = copy.copy( self.records[i][key])
        for key in ('connect',):
          subrec = []
          for line in self.records[i][key]:
            subrec.append( copy.copy( line))
          o.__dict__[ key] = subrec
      ## ATOM
      elif o.object_type == 'atom':
        for key in self.records[i].keys():
          if key == 'marks':
            o.__dict__[ key] = copy.copy( self.records[i][key])
          elif self.records[i][key] != o.__dict__[key]:
            changed = 1
            if key == 'name':
              # name is special
              if self.records[i]['type'] == 'text':
                o.set_name( self.records[i][key], interpret=0)
              else:
                o.set_name( self.records[i][key], interpret=1)
            #elif key == 'type':
            #  pass
            else:
              o.__dict__[key] = self.records[i][key]
      # VECTOR GRAPHICS
      elif o.object_type in data.vector_graphics_types:
        for key in ('line_color', 'area_color', 'line_width'):
          if self.records[i][key] != o.__dict__[key]:
            o.__dict__[key] = self.records[i][key]
        # polygon is special
        if o.object_type == 'polygon':
          for key in ('points',):
            if self.records[i][key] != o.__dict__[key]:
              o.__dict__[ key] = copy.copy( self.records[i][key])
        else:
          for key in ('coords',):
            o.__dict__[ key] = copy.copy( self.records[i][key])
        changed = 1 # redraws even if no change was made, should be fixed someday
      ## THE REST
      else:
        for key in self.records[i].keys():
          if self.records[i][key] != o.__dict__[key]:
            changed = 1
            o.__dict__[key] = self.records[i][key]
      ## COMMON
      if changed:
        if o.object_type == 'atom':
          [b.redraw() for b in o.molecule.atoms_bonds( o) if b not in deleted]
        elif o.object_type == 'point':
          o.arrow.redraw()  ## redraws arrow multiple times, should be fixed someday
        if o not in deleted:
          o.redraw()
      i += 1
    ## DELETED OBJECTS
      # deleted are known from the top of this def
    for o in deleted:
      if o.object_type == 'plus':
        self.paper.pluses.append( o)
      elif o.object_type == 'arrow':
        self.paper.arrows.append( o)
      elif o.object_type == 'text':
        self.paper.texts.append( o)
      elif o.object_type == 'molecule':
        self.paper.molecules.append( o)        
      elif o.object_type in data.vector_graphics_types:
        self.paper.vectors.append( o)
      if o.object_type != 'molecule':
        o.draw()
    ## ADDED OBJECTS
    added = misc.difference( previous.objects, self.objects)
    for o in added:
      if o.object_type == 'plus':
        self.paper.pluses.remove( o)
      elif o.object_type == 'arrow':
        self.paper.arrows.remove( o)
      elif o.object_type == 'text':
        self.paper.texts.remove( o)
      elif o.object_type == 'molecule':
        self.paper.molecules.remove( o)
      elif o.object_type in data.vector_graphics_types:
        self.paper.vectors.remove( o)
      if o.object_type != 'molecule':
        o.delete()

    self.paper.stack = copy.copy( self.stack)
    self.paper.add_bindings()

  def get_difference( self, previous):
    """obsolet method that finds difference between two state_records.
    Its too slow to be used in undo tracking so full state_record is used instead"""
    p_records = previous.records
    r_records = self.records
    p_objects = previous.objects
    r_objects = self.objects
    un_rec = undo_record( self.paper)
    for o in r_objects:
      if o not in p_objects:
        un_rec.added.append( o)
      else:
        a = p_records[ p_objects.index(o)]
        b = r_records[ r_objects.index(o)]
        if a != b:
          un_rec.modified.append( o)
          un_rec.modified_records.append( a)
    un_rec.deleted = misc.difference( p_objects, r_objects)
    return un_rec
    
  def _atom_rec( self, atom):
    """record for atoms"""
    rec = {}
    for key in ('x', 'y', 'z', 'pos', 'show', 'name', 'molecule', 'font_family',
                'font_size', 'charge', 'show_hydrogens', 'type', 'line_color', 'area_color'):
      rec[ key] = atom.__dict__[ key]
    for key in ('marks',):
      rec[ key] = copy.copy( atom.__dict__[ key])
    return rec

  def _mark_rec( self, m):
    """record for points"""
    rec = {}
    for key in ('x', 'y', 'auto'):
      rec[ key] = m.__dict__[ key]
    return rec


  def _bond_rec( self, bond):
    """record for bonds"""
    rec = {}
    for key in ('atom1', 'atom2', 'type', 'line_width', 'center', 'bond_width',
                'molecule', 'line_color','double_length_ratio'):
      rec[ key] = bond.__dict__[ key]
    return rec

  def _point_rec( self, point):
    """record for points"""
    rec = {}
    for key in ('x', 'y'):
      rec[ key] = point.__dict__[ key]
    return rec

  def _arrow_rec( self, arrow):
    """record for arrows"""
    rec = {}
    for key in ('pin', 'spline', 'line_width', 'line_color'):
      rec[ key] = arrow.__dict__[ key]
    for key in ('points',):
      rec[ key] = copy.copy( arrow.__dict__[ key])
    return rec

  def _plus_rec( self, plus):
    """record for plus"""
    rec = {}
    for key in ('x', 'y', 'font_size', 'font_family', 'line_color','area_color'):
      rec[ key] = plus.__dict__[ key]
    return rec

  def _text_rec( self, text):
    """record for texts"""
    rec = {}
    for key in ('x', 'y', 'text', 'font_size', 'font_family', 'line_color', 'area_color'):
      rec[ key] = text.__dict__[ key]
    return rec

  def _mol_rec( self, mol):
    """record for molecules"""
    rec = {}
    for key in ('name','id'):
      rec[ key] = mol.__dict__[ key]
    for key in ('atoms_map','bonds'):
      rec[ key] = copy.copy( mol.__dict__[ key])
    for key in ('connect',):
      subrec = []
      for line in mol.__dict__[ key]:
        subrec.append( copy.copy( line))
      rec[ key] = subrec
    return rec

  def _vector_rec( self, v):
    rec = {}
    for key in ('coords',):
      rec[ key] = copy.copy( v.__dict__[ key])
    for key in ('line_color', 'area_color', 'line_width'):
      rec[ key] = v.__dict__[ key]
    return rec

  def _polygon_rec( self, v):
    rec = {}
    for key in ('points',):
      rec[ key] = copy.copy( v.__dict__[ key])
    for key in ('line_color', 'area_color', 'line_width'):
      rec[ key] = v.__dict__[ key]
    return rec

  
